/******************************************************************************
** InterSense NavChip Developer Kit
**
** The InterSense NavChip Developer Kit sample code is considered proprietary
** and confidential. The NavChip Developer Kit sample code is available for
** use by customers who have signed the InterSense NDA and otherwise shall not
** be distributed without the written consent of InterSense Incorporated.
**
** Filename: NDK_SampleCode.c
** Modified: [see NDK_DATE below]
**
** Comments: The NavChip Developer Kit sample code provides generic C code
**           as an example of how to interact with the NavChip. The code is
**           organized into 3 layers. Each layer depends only on the layers
**           below it.
**
**           Application layer - NDK_SampleCode.c
**
**              This layer implements a text-based user interface using the
**              C run-time library to demonstrate the use of the underlying
**              layers.
**
**              The user interface can be either an interactive menu or
**              passive with minimal interaction. This is selectable at
**              compile time with the UI_MENU definition below.
**
**           NavChip interface layer - NavChip.h/.c
**
**              This layer is the middle layer and contains functions for
**              performing basic interactions with the NavChip. It
**              depends on the OS interface layer for all interactions
**              with the operating system.
**
**           OS interface layer - OS_Layer.h/OS_*.c
**
**              This is the bottom layer and encapsulates all OS-related
**              functions. It should be possible to port the sample code
**              to any OS by appropriately changing the implementation of
**              the functions in this layer with minimal effect on the upper
**              layers. OS_Win32.c is the implementation for Microsoft
**              WIN32 and OS_Linux.c is the implementation for Linux.
**
**           Public functions in the NavChip and OS interface layers are
**           documented in the respective header files.
**
**           Contact InterSense for technical support:
**           http://www.intersense.com
**           Tel: 781-541-6330
**
**           The sample source code in this project is provided for use by
**           InterSense customers with the limitations and restrictions as
**           defined above. It is supplied without any warranty or implied
**           suitability or fitness for any particular purpose.
******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "NavChip.h"

#define NDK_DATE "8 Jul 2015"

/* If defined, use interactive menu for user interface, else use
/* passive interface */
#define UI_MENU

/* Maximum refresh period (ms) for displaying NavChip data. This should be
/* high enough to allow data acquisition loop enough time to update the
/* display and keep up with data stream. If it is set too low (or data
/* rate is too high) the display may stall */
#define DISPLAY_PERIOD 100

/* Local variables */
static BOOL _connected = FALSE; /* Is NavChip connected? */
static int _port = 0;           /* NavChip port number (zero-based) */
static int _address = 0;        /* NavChip device address */
static int _rate = 0;           /* NavChip baud rate (bps) */

/* Local functions */
static void _runMenu( void );
static void _runPassive( void );
static void _showDeviceInfo( void );
static void _setPacketType( void );
static void _setDataRate( void );
static void _setBaudRate( void );
static void _displayData( void );
static int _getInt( void );

/*****************************************************************************/

int main( void )
{
#ifdef UI_MENU

    /* Run interactive menu interface */
    _runMenu();

#else

    /* Run passive interface */
    _runPassive();

    /* Pause before exit */
    printf( "\nPress return to exit\n" );
    _getInt();

#endif

    return 0;
}

/******************************************************************************
** _runMenu - Dispatches interactive menu
**
** Inputs/Outputs/Returns:
**   none
******************************************************************************/
static void _runMenu( void )
{
    int choice;

    while (TRUE)
    {
        /* Clear the screen */
        osConsoleClear();

        /* Display header information */
        printf( "InterSense NavChip Developer Kit - " NDK_DATE "\n\n" );
        printf( "http://www.intersense.com\n" );
        printf( "Tel: 781-541-6330\n\n" );

        /* Show NavChip comm info */
        printf( "===============================\n" );
        if( _connected )
        {
            printf( "NavChip    : Connected\n" );
            printf( "Port Number: %d\n", _port + 1 );
            printf( "Baud Rate  : %d\n", _rate );
            printf( "Address    : %d\n", _address );
        }
        else
        {
            printf( "NavChip    : Not Connected\n" );
            printf( "Port Number: \n" );
            printf( "Baud Rate  : \n" );
            printf( "Address    : \n" );
        }
        printf( "===============================\n\n" );

        /* Show the user options */
        printf( "Select one of the following options to continue:\n\n" );
        printf( "  1. Autodetect and Connect\n" );
        printf( "  2. Show Device Info\n" );
        printf( "  3. Set Packet Type\n" );
        printf( "  4. Set Data Rate\n" );
        printf( "  5. Set Baud Rate\n" );
        printf( "  6. Display Streaming Data\n" );
        printf( "  7. Exit\n\n" );
        printf( "Choice: " );

    	/* Get user input */
        choice = _getInt();

    	/* Handle user input here */
        switch( choice )
        {
		/* Autodetect and connect to NavChip */
		case 1:
            printf( "Autodetecting...\n\n" );

            _connected = ncOpen( &_port, &_address, &_rate );

            if (_connected)
            {
                printf( "NavChip found\n" );
            }
            else
            {
                printf( "No NavChips found\n" );
            }
            break;

		/* Read device information and display it */
		case 2:
            if (_connected)
            {
                _showDeviceInfo();
            }
            else
            {
                printf( "Please connect first...\n" );
            }
            break;
		
		/* Set packet type */
		case 3:
            if( _connected )
            {
                _setPacketType();
            }
            else
            {
                printf( "Please connect first...\n" );
            }
            break;
		
		/* Set data rate */
		case 4:
            if( _connected )
            {
                _setDataRate();
            }
            else
            {
                printf( "Please connect first...\n" );
            }
            break;
		
		/* Set baud rate */
		case 5:
            if( _connected )
            {
                _setBaudRate();
            }
            else
            {
                printf( "Please connect first...\n" );
            }
            break;
		
		/* Display packet data */
		case 6:
            if( _connected )
            {
                _displayData();
            }
            else
            {
                printf( "Please connect first...\n" );
            }
            break;

		/* Exit */
		case 7:
            ncClose();
            return;

		/* Invalid input */
		default :
            printf( "Invalid selection...\n" );
    	}

    	/* Pause after completion of task */
    	printf( "\nPress return to continue\n" );
        _getInt();
    }
}

/******************************************************************************
** _showDeviceInfo - Displays some NavChip information
**
** Inputs/Outputs/Returns:
**   none
******************************************************************************/
static void _showDeviceInfo( void )
{
    INT32 serialNum;
    INT32 navChipType;
    INT32 firmwareRev;
    float dataRate;
    INT32 packetType;
    int syncEnabled;
    int streamOnPowerup;
    int userBoresightSet;
    INT32 calRev;
    INT16 temp;
    float temperature;
    UINT8 value[4];

    printf( "Reading device info...\n\n" );

    /* Get register values and parse them as needed */

    if( !ncGetRegister( NC_REG_SERIAL_NUMBER, 3, value ) )
    {
        printf( "ERROR: Failed to get serial number\n" );
        return;
    }
    value[3] = 0;
    serialNum = ncLittleEndian32( value );

    if( !ncGetRegister( NC_REG_NAVCHIP_TYPE, 1, value ) )
    {
        printf( "ERROR: Failed to get NavChip type\n" );
        return;
    }
    navChipType = value[0];

    if( !ncGetRegister( NC_REG_FIRMWARE_VERSION, 2, value ) )
    {
        printf( "ERROR: Failed to get firmware version\n" );
        return;
    }
    firmwareRev = ncLittleEndian16( value );

    if( !ncGetRegister( NC_REG_TEMPERATURE, 2, value ) )
    {
        printf( "ERROR: Failed to get temperature\n" );
        return;
    }
    temp = ncLittleEndian16( value );
    temp = temp >> 4; /* value is upper 12 bits (right shift preserves sign bit) */
    temperature = temp * NC_SCALE_TEMP;

    if( !ncGetRegister( NC_REG_DATA_RATE_DIVISOR, 1, value ) )
    {
        printf( "ERROR: Failed to get data rate divisor\n" );
        return;
    }
    dataRate = ( (float) NC_MAX_DATA_RATE ) / value[0];

    if( !ncGetRegister( NC_REG_PACKET_TYPE, 1, value ) )
    {
        printf( "ERROR: Failed to get packet type\n" );
        return;
    }
    packetType = value[0];

    if( !ncGetRegister( NC_REG_ENABLE_FLAGS, 1, value ) )
    {
        printf( "ERROR: Failed to get enable flags\n" );
        return;
    }
    streamOnPowerup  = ( value[0] & 0x01 ) != 0;
    syncEnabled      = ( value[0] & 0x02 ) != 0;
    userBoresightSet = ( value[0] & 0x04 ) != 0;

    if( !ncGetRegister( NC_REG_CAL_REVISION, 2, value ) )
    {
        printf( "ERROR: Failed to get enable flags\n" );
        return;
    }
    calRev = ncLittleEndian16( value );

    /* Show device information */
    printf( "[ NAVCHIP DEVICE INFO ]\n");
    printf( "=======================\n");
    printf( "Serial Number      : %d\n", (int) serialNum );
    printf( "NavChip Type       : %d\n", (int) navChipType );
    printf( "Firmware Version   : %d\n", (int) firmwareRev );
    printf( "Temperature (deg C): %.2f\n", (float) temperature );
    printf( "Data Rate (Hz)     : %.2f\n", (float) dataRate );
    printf( "Packet Type        : %d\n", (int) packetType );
    printf( "Sync Enabled       : %d\n", (int) syncEnabled );
    printf( "Stream on Powerup  : %d\n", (int) streamOnPowerup );
    printf( "User Boresight     : %d\n", (int) userBoresightSet );
    printf( "Cal Revision       : %d\n", (int) calRev );
}

/******************************************************************************
** _setPacketType - Set NavChip packet type
**
** Inputs/Outputs/Returns:
**   none
******************************************************************************/
static void _setPacketType( void )
{
    UINT8 reg;
    int value;

    if( !ncGetRegister( NC_REG_PACKET_TYPE, 1, &reg ) )
    {
        printf( "ERROR: Failed to get packet type\n" );
        return;
    }
    value = reg;
    printf( "Current packet type  : %d\n", value );
    printf( "New packet type (1-9): " );
    value = _getInt();
    
    if( value < 1 || value > 9 )
    {
        printf( "ERROR: Entry invalid\n" );
        return;
    }

    if( !ncSetRegister1( NC_REG_PACKET_TYPE, (UINT8) value ) )
    {
        printf( "ERROR: Failed to set packet type\n" );
        return;
    }

    printf( "Packet type set successfully\n" );
}    

/******************************************************************************
** _setDataRate - Set NavChip data rate
**
** Inputs/Outputs/Returns:
**   none
******************************************************************************/
static void _setDataRate( void )
{
    UINT8 reg;
    int value;

    if( !ncGetRegister( NC_REG_DATA_RATE_DIVISOR, 1, &reg ) )
    {
        printf( "ERROR: Failed to get data rate divisor\n" );
        return;
    }
    value = reg;
    printf( "Current data rate divisor   : %d\n", value );
    printf( "New data rate divisor (1-10): " );
    value = _getInt();
    
    if( value < 1 || value > 10 )
    {
        printf( "ERROR: Entry invalid\n" );
        return;
    }

    if( !ncSetRegister1( NC_REG_DATA_RATE_DIVISOR, (UINT8) value ) )
    {
        printf( "ERROR: Failed to set data rate divisor\n" );
        return;
    }

    printf( "Data rate divisor set successfully\n" );
}    

/******************************************************************************
** _setBaudRate - Set NavChip baud rate and change port speed to match
**
** Inputs/Outputs/Returns:
**   none
******************************************************************************/
static void _setBaudRate( void )
{
    UINT8 reg;
    int value;
    int baudRate;

    if( !ncGetRegister( NC_REG_BAUD_RATE_DIVISOR, 1, &reg ) )
    {
        printf( "ERROR: Failed to get baud rate divisor\n" );
        return;
    }
    value = reg;
    printf( "Current baud rate divisor   : %d\n", value );
    printf( "New baud rate divisor (1-24): " );
    value = _getInt();
    
    if( value < 1 || value > 24 )
    {
        printf( "ERROR: Entry invalid\n" );
        return;
    }

    if( !ncSetRegister1( NC_REG_BAUD_RATE_DIVISOR, (UINT8) value ) )
    {
        printf( "ERROR: Failed to set baud rate divisor\n" );
        return;
    }
    baudRate = NC_MAX_BAUD_RATE / value;

    if( !osComPortSetRate( _port, baudRate ) )
    {
        printf( "ERROR: Failed to set baud rate\n" );
        return;
    }
    _rate = baudRate;

    printf( "Baud rate set successfully\n" );
}

/******************************************************************************
** _displayData - Turns data streaming on and displays data in form
**                format. Packet errors indicate timeouts and invalid
**                packets. Sequence errors indicate skip in packet ID. Both
**                are normally zero
**
** Inputs/Outputs/Returns:
**   none
******************************************************************************/
static void _displayData( void )
{
    BOOL ready;
    ncPacketDataType data;
    int pktCount = 0;
    int peCount = 0;
    int seCount = 0;
    int i, j;
    float deltaTheta[3];
    float deltaV[3];
    float mag[3];
    float euler[3];
    float quat[4];
    float rot[3][3];
    UINT32 t0, tElapsed;
    UINT8 prevPktId = 0;

    memset( &data, 0, sizeof( data ) );

    for( i = 0; i < 3; i++ )
    {
        deltaTheta[i] = 0;
        deltaV[i] = 0;
        mag[i] = 0;
    }

    if( !ncStartStreaming() )
    {
        printf( "ERROR: Failed to start streaming\n" );
        return;
    }

    osConsoleClear();
    t0 = osTimeGet();

    /* Data acquisition loop (press any key to stop) */
    while( !osKbhit() )
    {
        /* Check if packet has arrived and read it */
        if( !ncReadPacket( &ready ) )
        {
            /* Packet error */
            peCount++;
        }

        /* Retrieve data if packet did arrive */
        if( ready )
        {
            pktCount++;

            if( !ncGetData( &data ) )
            {
                printf( "ERROR: Failed to get data\n" );
                break;
            }

            /* Validate packet ID */
            if( pktCount > 1 )
            {
                if( !ncValidatePacketId( prevPktId, data.packetId ) )
                {
                    /* Sequence error */
                    seCount++;
                }
            }
            prevPktId = data.packetId;
        }

        tElapsed = osTimeGetElapsed( t0 );

        /* Update display at display rate (do this only during idle
        /* time while waiting for packet so that data acquisition
        /* gets priority over display update) */
        if( !ready && tElapsed >= DISPLAY_PERIOD )
        {
            osConsoleHome();

            /* Scale data to floating point values */
            for( i = 0; i < 3; i++ )
            {
                if( data.packetType < 3 )
                {
                    deltaTheta[i] = data.deltaTheta[i] * NC_SCALE_DTC;
                    deltaV[i] = data.deltaV[i] * NC_SCALE_DVC;
                }
                else
                {
                    deltaTheta[i] = data.deltaTheta[i] * NC_SCALE_DT;
                    deltaV[i] = data.deltaV[i] * NC_SCALE_DV;
                }
                euler[i] = data.euler[i] * NC_SCALE_EULER;
            }
            for( i = 0; i < 4; i++ )
            {
                quat[i] = data.quat[i] * NC_SCALE_QUAT;
            }
            for( i = 0; i < 3; i++ )
            {
                for( j = 0; j < 3; j++ )
                {
                    rot[i][j] = data.rot[i][j] * NC_SCALE_ROT;
                }
            }

            /* Parse mag value according to index */
            i = data.flags & NC_FLAG_I;

            if( i > 0 )
            {
                mag[i-1] = data.mag * NC_SCALE_MAG;
            }

            printf( "[ NAVCHIP DATA ]\n" );
            printf( "================\n" );
            printf( "Display Status    : %8s\n", "active" );
            printf( "Packets Received  : %8d\n", pktCount );
            printf( "Packet Errors     : %8d\n", peCount );
            printf( "Sequence Errors   : %8d\n", seCount );
            printf( "================\n" );
            printf( "Packet Type       : %8d\n", (int) data.packetType );
            printf( "Packet ID         : %8d\n", (int) data.packetId );
            printf( "Flags             :     %#04x\n", (unsigned) data.flags );
            printf( "Timestamp (us)    : %8d\n", (int) data.timestamp );
            printf( "Delta Theta (mrad): %8.3f %8.3f %8.3f\n",
                deltaTheta[0], deltaTheta[1], deltaTheta[2] );
            printf( "Delta V  (m/s)    : %8.3f %8.3f %8.3f\n",
                deltaV[0], deltaV[1], deltaV[2] );
            printf( "Mags (G)          : %8.3f %8.3f %8.3f\n",
                mag[0], mag[1], mag[2] );
            printf( "Euler Angles (rad): %8.3f %8.3f %8.3f\n",
                euler[0], euler[1], euler[2] );
            printf( "Quaternion        : %8.3f %8.3f %8.3f %8.3f\n",
                quat[0], quat[1], quat[2], quat[3] );
            printf( "Rot[0][]          : %8.3f %8.3f %8.3f\n",
                rot[0][0], rot[0][1], rot[0][2] );
            printf( "Rot[1][]          : %8.3f %8.3f %8.3f\n",
                rot[1][0], rot[1][1], rot[1][2] );
            printf( "Rot[2][]          : %8.3f %8.3f %8.3f\n",
                rot[2][0], rot[2][1], rot[2][2] );
            printf( "\nPress any key to stop\n" );
            t0 = osTimeGet();
        }
        /* If idle, sleep to free CPU */
        else if( !ready )
        {
            osTimeWait( 1 );
        }
        /* This condition means that process is not running fast
        /* enough to keep up with packet data rate. Try decreasing
        /* data rate or increasing display period */
        else if( tElapsed >= DISPLAY_PERIOD * 5 )
        {
            osConsoleHome();
            printf( "\n\n" );
            printf( "Display Status    : %8s\n", "stalled" );
            printf( "\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n" );
            t0 = osTimeGet();
        }
    }

    if( !ncStopStreaming() )
    {
        printf( "ERROR: Failed to stop streaming\n" );
        return;
    }
}

/******************************************************************************
** _runPassive - Runs passive interface with minimal user interaction
**
** Inputs/Outputs/Returns:
**   none
******************************************************************************/
static void _runPassive( void )
{
    INT32 serialNum;
    UINT8 value[4];

    printf( "InterSense NavChip Developer Kit - " NDK_DATE "\n" );

    /* Connect */
    printf( "Autodetecting...\n" );
    _connected = ncOpen( &_port, &_address, &_rate );

    if( !_connected )
    {
    	printf( "No NavChips found\n" );
        return;
    }
    printf( "NavChip found on port %d, address %d, baud %d\n",
        _port, _address, _rate );

    /* Get serial number */
    if( !ncGetRegister( NC_REG_SERIAL_NUMBER, 3, value ) )
    {
        printf( "ERROR: Failed to get serial number\n" );
        return;
    }
    value[3] = 0;
    serialNum = ncLittleEndian32( value );
    printf( "NavChip serial number: %d\n", serialNum );

    /* Set packet type (3=default) */
    if( !ncSetRegister1( NC_REG_PACKET_TYPE, 3 ) )
    {
        printf( "ERROR: Failed to set packet type\n" );
        return;
    }

    /* Set data rate divisor (5=default) */
    if( !ncSetRegister1( NC_REG_DATA_RATE_DIVISOR, 5 ) )
    {
        printf( "ERROR: Failed to set data rate divisor\n" );
        return;
    }

     printf( "Press return to begin display\n" );
    _getInt();

    /* Display streaming data */
    _displayData();

    ncClose();
}

/******************************************************************************
** _getInt - Get integer from standard input (also can be used to wait for
**           return key)
**
** Inputs/Outputs:
**   none
**
** Returns:
**   value
******************************************************************************/
static int _getInt( void )
{
    char s[20];

    if( fgets( s, sizeof( s ), stdin ) == NULL )
    {
        return 0;
    }

    /* Flush any remaining bytes */
    fflush( stdin );

    return atoi( s );
}
