/******************************************************************************
** InterSense NavChip Developer Kit
**
** The InterSense NavChip Developer Kit sample code is considered proprietary
** and confidential. The NavChip Developer Kit sample code is available for
** use by customers who have signed the InterSense NDA and otherwise shall not
** be distributed without the written consent of InterSense Incorporated.
**
** Filename: OS_Win32.c
** Modified: 14 Dec 2009
**
** Comments: Implementation for WIN32.	
**           See also osInterface.h for public function descriptions.
**
**           The sample source code in this project is provided for use by
**           InterSense customers with the limitations and restrictions as
**           defined above. It is supplied without any warranty or implied
**           suitability or fitness for any particular purpose.
******************************************************************************/

#ifndef WIN32
#error This module for use with WIN32 only
#endif

#include <windows.h>
#include <stdio.h>
#include <conio.h>
#include "OS_Layer.h"

/******************************************************************************
***************************** Serial port access ******************************
******************************************************************************/

#define MAX_COM_PORTS ( OS_MAX_PORT_NUMBER + 1 )

typedef struct
{
    char name[100];
    HANDLE handle;
}
ComPort;

static ComPort _comPort[MAX_COM_PORTS];

/******************************************************************************
** _comPortGet - Get port object corresponding to given port number
**
** Inputs:
**   portNumber - zero-based port number
**
** Outputs:
**   none
**
** Returns:
**   pointer to port object (NULL if none)
******************************************************************************/
static ComPort *_comPortGet( int portNumber )
{
    ComPort *port = NULL;
    BOOL ok = portNumber >=0 && portNumber < MAX_COM_PORTS;

    if( !ok )
    {
        osErrorDisplay( "Supplied port number is out of range" );
    }
    else
    {
        port = _comPort + portNumber;
    }
    return port;
}

/*****************************************************************************/

BOOL osComPortOpen( int portNumber )
{
    ComPort *port;
    COMMTIMEOUTS tmo;

    /* Get port */
    if( !( port = _comPortGet( portNumber ) ) )
    {
        return FALSE;
    }

    osComPortClose( portNumber );

    sprintf( port->name, "\\\\.\\COM%d", portNumber + 1 );

    /* Open COM port (non-overlapped for simplicity) */
    port->handle = CreateFile( port->name,
        GENERIC_READ | GENERIC_WRITE, 0, 0, OPEN_EXISTING,
        FILE_ATTRIBUTE_NORMAL, 0 );

    /* Port cannot be opened? */
    if( port->handle == INVALID_HANDLE_VALUE )
    {
#ifdef _VERBOSE
        osErrorDisplay( "Unable to open serial port" );
        osErrorDisplay( NULL );
#endif
        port->handle = NULL;
        return FALSE;
    }

    if ( !GetCommTimeouts( port->handle, &tmo ) )
    {
        osErrorDisplay( "Failed to get timeouts for serial port" );
        osErrorDisplay( NULL );
        osComPortClose( portNumber );
        return FALSE;
    }

    /* Set read timeouts for non-blocking access */
    tmo.ReadIntervalTimeout = MAXDWORD;
    tmo.ReadTotalTimeoutMultiplier = 0;
    tmo.ReadTotalTimeoutConstant = 0;

    /* Set write timeouts for 10 ms per byte + 100 ms (should be ample) */
    tmo.WriteTotalTimeoutMultiplier = 10;
    tmo.WriteTotalTimeoutConstant = 100;

    if( !SetCommTimeouts( port->handle, &tmo ) )
    {
        osErrorDisplay( "Failed to set timeouts for serial port" );
        osErrorDisplay( NULL );
        osComPortClose( portNumber );
        return FALSE;
    }

    return TRUE;
}

/*****************************************************************************/

BOOL osComPortClose( int portNumber )
{
    ComPort *port;

    if( !( port = _comPortGet( portNumber ) ) )
    {
        return FALSE;
    }

    if ( port->handle != NULL )
    {
        CloseHandle( port->handle );
        port->handle = NULL;
    }
    return TRUE;
}

/*****************************************************************************/

BOOL osComPortFlush( int portNumber )
{
    ComPort *port;

    if( !( port = _comPortGet( portNumber ) ) )
    {
        return FALSE;
    }

    if ( port->handle == NULL )
    {
        osErrorDisplay( "Port is not open" );
        return FALSE;
    }

    if( !PurgeComm( port->handle, PURGE_RXCLEAR ) )
    {
        osErrorDisplay( "Failed to flush serial port" );
        osErrorDisplay( NULL );
        return FALSE;
    }

    return TRUE;
}

/*****************************************************************************/

BOOL osComPortSetRate( int portNumber, int rate )
{
    ComPort *port;
    DCB dcb;

    if( !( port = _comPortGet( portNumber ) ) )
    {
        return FALSE;
    }

    if ( port->handle == NULL )
    {
        osErrorDisplay( "Port is not open" );
        return FALSE;
    }

    FillMemory( &dcb, sizeof( dcb ), 0 );
    dcb.DCBlength = sizeof( dcb );

    if ( !GetCommState( port->handle, &dcb ) )
    {
        osErrorDisplay( "Failed to get state for serial port" );
        osErrorDisplay( NULL );
        return FALSE;
    }
  
    dcb.DCBlength = sizeof(dcb);
    dcb.BaudRate = rate;
    dcb.fParity = FALSE;
    dcb.Parity = NOPARITY;
    dcb.ByteSize = 8;  
    dcb.StopBits = ONESTOPBIT;
    dcb.fDsrSensitivity = FALSE;
    dcb.fTXContinueOnXoff = FALSE;
    dcb.fNull = FALSE;
    dcb.fBinary = TRUE;
    dcb.fAbortOnError = FALSE;
    dcb.fOutxCtsFlow = FALSE;
    dcb.fOutxDsrFlow = FALSE;    
    dcb.fOutX = FALSE;
    dcb.fInX = FALSE;
    dcb.fRtsControl = RTS_CONTROL_ENABLE;
    dcb.fDtrControl = DTR_CONTROL_ENABLE;

    if ( !SetCommState( port->handle, &dcb ) )
    {
#ifdef _VERBOSE
        osErrorDisplay( "Failed to set state for serial port" );
        osErrorDisplay( NULL );
#endif
        return FALSE;
    }

    return TRUE;
}

/*****************************************************************************/

BOOL osComPortRead( int portNumber, UINT8 buffer[], int reqCount, int *rcvCount )
{
    ComPort *port;

    if( !( port = _comPortGet( portNumber ) ) )
    {
        return FALSE;
    }

    if ( port->handle == NULL )
    {
        osErrorDisplay( "Port is not open" );
        return FALSE;
    }

    /* Read specified number of bytes */
    if( !ReadFile( port->handle, buffer, reqCount, rcvCount, NULL ) )
    {
        osErrorDisplay( "Read from serial port failed" );
        osErrorDisplay( NULL );
        return FALSE;
    }

    return TRUE;
}

/*****************************************************************************/

BOOL osComPortWrite( int portNumber, UINT8 buffer[], int count )
{
    ComPort *port;
    int writeCount;

    if( !( port = _comPortGet( portNumber ) ) )
    {
        return FALSE;
    }

    if ( port->handle == NULL )
    {
        osErrorDisplay( "Port is not open" );
        return FALSE;
    }

    if( !WriteFile( port->handle, buffer, count, &writeCount, NULL ) )
    {
        osErrorDisplay( "Write to serial port failed" );
        osErrorDisplay( NULL );
        return FALSE;
    }

    return TRUE;
}


/******************************************************************************
******************************* Timer access **********************************
******************************************************************************/

UINT32 osTimeGet( void )
{
    static BOOL _timeInitialized = FALSE;
    static LARGE_INTEGER _timeTicksPerSec, _timeStartTicks;
    LARGE_INTEGER ticks;
    UINT32 time;

    /* Get initial time on the first time through */
    if( !_timeInitialized )
    {
        QueryPerformanceFrequency( &_timeTicksPerSec );
        QueryPerformanceCounter( &_timeStartTicks );
        _timeInitialized = TRUE;
    }

    /* Subtract the initial time from current time.
    /* Output value (ms) will overflow after 50 days */
    QueryPerformanceCounter( &ticks );
    time = (UINT32) ( ( ticks.QuadPart - _timeStartTicks.QuadPart ) /
        ( _timeTicksPerSec.QuadPart / 1000 ) );

    return time;
}

/*****************************************************************************/

UINT32 osTimeGetElapsed( UINT32 t0 )
{
    UINT32 t1 = osTimeGet();

    /* If timer overflows, set first value to zero (crude but safe) */
    if( t1 < t0 )
    {
        t0 = 0;
    }

    return t1 - t0;
}

/*****************************************************************************/

BOOL osTimeWait( UINT32 delay )
{
    Sleep( delay );
    return TRUE;
}


/******************************************************************************
*********************************** Other *************************************
******************************************************************************/

/******************************************************************************
** _consoleGet - Get handle to console
**
** Inputs/Outputs:
**   none
**
** Returns:
**   handle (NULL on error)
******************************************************************************/
static HANDLE _consoleGet( void )
{
    static HANDLE h = NULL;

    if( h == NULL )
    {
        h = CreateFile( "CONOUT$", GENERIC_WRITE | GENERIC_READ,
            FILE_SHARE_READ | FILE_SHARE_WRITE,
            0L, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0L );
    }
    return h;
}

/*****************************************************************************/

BOOL osConsoleClear( void )
{ 
    HANDLE hConsole;
    CONSOLE_SCREEN_BUFFER_INFO csbi;
    COORD coordScreen = { 0, 0 }; /* Cursor home */

    if( ( hConsole = _consoleGet() ) == (HANDLE) -1 )
    {
        osErrorDisplay( "Unable to open console" );
        osErrorDisplay( NULL );
        return FALSE;
    } 
    /* Get the number of character cells in the current buffer */
    if( !GetConsoleScreenBufferInfo( hConsole, &csbi ) )
    {
        osErrorDisplay( "Console access failed" );
        return FALSE;
    }

    {
        DWORD cCharsWritten;
        DWORD dwConSize;

        dwConSize = csbi.dwSize.X * csbi.dwSize.Y; 

        /* Fill the entire screen with blanks */
        if( !FillConsoleOutputCharacter( hConsole, (TCHAR) ' ', 
            dwConSize, coordScreen, &cCharsWritten ) ||
            /* Get the current text attribute */
            !GetConsoleScreenBufferInfo( hConsole, &csbi ) ||
            /* Now set the buffer's attributes accordingly */
            !FillConsoleOutputAttribute(hConsole, csbi.wAttributes,
            dwConSize, coordScreen, &cCharsWritten ) )
        {
            osErrorDisplay( "Console access failed" );
            return FALSE;
        }
    }

    /* Home cursor */
    if( !SetConsoleCursorPosition( hConsole, coordScreen ) )
    {
        osErrorDisplay( "Console access failed" );
        return FALSE;
    }

    return TRUE;
}

/*****************************************************************************/

BOOL osConsoleHome( void )
{ 
    HANDLE hConsole;
    COORD coordScreen = { 0, 0 }; /* Cursor home */

    if( ( hConsole = _consoleGet() ) == (HANDLE) -1 )
    {
        osErrorDisplay( "Unable to open console" );
        osErrorDisplay( NULL );
        return FALSE;
    } 

    if( !SetConsoleCursorPosition( hConsole, coordScreen ) )
    {
        osErrorDisplay( "Console access failed" );
        return FALSE;
    }

    return TRUE;
}

/*****************************************************************************/

BOOL osKbhit( void )
{
    return kbhit();
}

/*****************************************************************************/

BOOL osErrorDisplay( const char *message )
{
    /* Display supplied message, if any */
    if( message != NULL )
    {
        printf( message );
        printf( "\n" );
    }
    /* Get system error if no message supplied */
    else
    {
        char *msg;

        FormatMessage( 
            FORMAT_MESSAGE_ALLOCATE_BUFFER | 
            FORMAT_MESSAGE_FROM_SYSTEM     |
            FORMAT_MESSAGE_IGNORE_INSERTS,
            NULL, 
            GetLastError(),
            MAKELANGID( LANG_NEUTRAL, SUBLANG_DEFAULT ),
            (LPTSTR) &msg, 0, NULL );

        printf( msg );
    }

    return TRUE;
}
 